/*
  ParseSensor.cpp
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include "ParseSensor.h"
#include <SafeString.h>

static Print* debugPtr = NULL;
void setParseSensorDebugOut(Print* outPtr) {
  debugPtr = outPtr;
}

bool decode(SafeString &sfNameIn);

static const size_t MAX_BLE_NAME_SIZE = 32;
static const char T_DEVICE[] = "T_"; // T_ devices have temp only (in degC)
static const char H_DEVICE[] = "H_"; // H_ devices have temp(degC),RH(%)
static const char W_DEVICE[] = "W_"; // W_ devices have temp(degC),RH(%),AirPressure(hPa)
static const char WL_DEVICE[] = "WL_"; // WL_ devices have temp(degC),RH(%),AirPressure(hPa),Lux
cSF(sfDevice, MAX_BLE_NAME_SIZE);
cSF(sf_TempC, MAX_BLE_NAME_SIZE);
cSF(sf_RH, MAX_BLE_NAME_SIZE);
cSF(sf_hPa, MAX_BLE_NAME_SIZE);
cSF(sf_lx, MAX_BLE_NAME_SIZE);
static float tempC =  1000;// invalid;
static float RH = 0;
static float hPa = 0;
static float lx = 0.01;  // will always be this if not lux reading decoded

static volatile float luxReading  = -100.0; // not set yet
static volatile bool haveNewLuxReading = false;

float getLuxReading() {
  return luxReading; // pickup volatile and return it.
}

cSF(sfSensorName, MAX_BLE_NAME_SIZE);

void setLuxSensorName(const char* luxSensorName) {
  sfSensorName = luxSensorName;
}

bool checkValidName(SafeString &sfName) {
  return (sfName.startsWith(sfSensorName)); // only WL_ device
}

bool parseDeviceName(const char* name) {
  cSF(sfName, strlen(name));
  sfName = name;
  sfName.trim();
  if (!checkValidName(sfName)) {
    return false;
  }
  // else
  if (debugPtr) {
    debugPtr->print("Found Device: ");
    debugPtr->println(name);
  }
  if (decode(sfName)) {
    luxReading = lx;
    haveNewLuxReading = true;
    if (debugPtr) {
      debugPtr->print("Lux value: ");
      debugPtr->println(luxReading);
    }
    return true;
  } else {
    return false;
  }
  return false;
}

bool decode(const char* nameIn) {
  cSF(sfNameIn, strlen(nameIn));
  sfNameIn = nameIn;
  return decode(sfNameIn);
}

bool decode(SafeString &sfNameIn) {
  sfNameIn.trim();
  if (!checkValidName(sfNameIn)) {
    return false;
  }
  cSF(sfName, sfNameIn.length());
  sfName = sfNameIn;
  // no need to clear the previous tokens as nextToken does that
  if (sfName.startsWith(T_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pick up temp
    if (debugPtr) {
      debugPtr->print(sfDevice); debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->println("C");
    }
    // convert Temp
    if (!sf_TempC.toFloat(tempC)) {
      return false; // last tempC unchanged
    }
    tempC = tempC; // set global
  } else if (sfName.startsWith(H_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    if (debugPtr) {
      debugPtr->print(sfDevice);
      debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->print("C");
      debugPtr->print(" RH:"); debugPtr->print(sf_RH); debugPtr->print("%");
      debugPtr->println();
    }
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float, last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float, last RH unchanged
    }
  } else if (sfName.startsWith(W_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    sfName.nextToken(sf_hPa, ','); // pickup Air Pressure
    if (debugPtr) {
      debugPtr->print(sfDevice);
      debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->print("C");
      debugPtr->print(" RH:"); debugPtr->print(sf_RH); debugPtr->print("%");
      debugPtr->print(" Barometer:"); debugPtr->print(sf_hPa); debugPtr->print(" hPa");
      debugPtr->println();
    }
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float, last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float, last RH unchanged
    }
    if (!sf_hPa.toFloat(hPa)) {
      return false; // not a valid float, last hPA unchanged
    }
  } else if (sfName.startsWith(WL_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    sfName.nextToken(sf_hPa, ','); // pickup Air Pressure
    sfName.nextToken(sf_lx, ','); // pickup Air Pressure
    if (debugPtr) {
      debugPtr->print(sfDevice);
      debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->print("C");
      debugPtr->print(" RH:"); debugPtr->print(sf_RH); debugPtr->print("%");
      debugPtr->print(" Barometer:"); debugPtr->print(sf_hPa); debugPtr->print(" hPa");
      debugPtr->print(" Lux:"); debugPtr->print(sf_lx); debugPtr->print(" lx");
      debugPtr->println();
    }
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float,  last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float,  last RH unchanged
    }
    if (!sf_hPa.toFloat(hPa)) {
      return false; // not a valid float,  last hPA unchanged
    }
    if (!sf_lx.toFloat(lx)) {
      return false; // not a valid float,  last lx unchanged
    }
  } else {
    if (debugPtr) {
      debugPtr->print("Error: Device not supported -- "); debugPtr->println(sfName);
    }
    return false;
  }
  return true;
}
