/*
  PowerSwitchMsgParser.cpp
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include <SafeString.h>
#include "PowerSwitchMsgParser.h"

static Print* debugPtr = NULL;
void setPowerSwitchMsgParserDebugOut(Print* outPtr) {
  debugPtr = outPtr;
}

static bool haveNewPowerSwitchStatus = false;
static bool powerIsOn = false;
static volatile float batteryVolts = 0.0; // if not read or invalid
static int adcReading = 0;

int getADCreading() {
  // returns last read PowerSwitch adc reading 1023 == 3V
  return adcReading;
}

bool isPowerSwitchOn() {
  // returns true if last PowerSwitch msg had PowerSwitch On
  return powerIsOn;
}

// local override when sending cmd
// call this with the expected result,
// next reading will set the correct value
void setPowerIsOn(bool flag) {
  powerIsOn = flag;
}

// this volatile is loaded by the UART_Client call back on its thread
volatile char powerSwitchMsg[MAX_PFOD_MSG_LEN] = "";
static char deviceMsg[MAX_PFOD_MSG_LEN] = "";

// call this on the loop() thread
bool parsePowerSwitchMsg() {
  if (*powerSwitchMsg == '\0') {
    return false; // no new msg
  }
  //else parse new msg
  volatile char *msg = powerSwitchMsg;
  char *psMsg = deviceMsg;

  while (*msg) {
    *psMsg = *msg;
    psMsg++;
    msg++;
  }
  *psMsg = '\0';
  *powerSwitchMsg = '\0'; // clear for next msg

  cSFA(sfDeviceMsg, deviceMsg);
  // looking for Power is ON / OFF
  sfDeviceMsg.replace('\r', ' ');
  sfDeviceMsg.replace('\n', ' ');
  sfDeviceMsg.replace("  ", " ");
  if (debugPtr) {
    //    debugPtr->println(sfDeviceMsg);
  }
  sfDeviceMsg.toLowerCase();
  if (sfDeviceMsg.indexOf("power is on") >= 0) {
    powerIsOn = true;
  } else if (sfDeviceMsg.indexOf("power is off") >= 0) {
    powerIsOn = false;
  } else {
    if (debugPtr) {
      debugPtr->println("Error no power status");
    }
    return false; // let connection timeout on error
  }
  // look for battery voltage reading 1023 == 3V
  char fieldStart[] = "|!c`";
  int idx = sfDeviceMsg.indexOf(fieldStart); // sfDeviceMsg is now in lower case, actual msg is |!C`
  if (idx >= 0) {
    // pick up adc reading following the `
    cSF(sfToken, 10); // number should only be 4 digits max
    sfDeviceMsg.stoken(sfToken, idx + strlen(fieldStart) - 1, "0123456789", true, false);
    // idx+strlen(fieldStart)-1 => start scanning from ` ,  true => skips one non-delimiter to start, false => uses delimiters at valid token chars,
    adcReading = 0;
    if (sfToken.toInt(adcReading)) {
      // found valid int.
      batteryVolts = (3.0 * adcReading) / 1023.0;
    } else {
      batteryVolts = 0.0; // invalid
    }
  } else { // field not found
    batteryVolts = 0.0; // invalid
  }
  if (debugPtr) {
    debugPtr->print("Power is "); debugPtr->print(powerIsOn ? "ON" : "OFF");
    debugPtr->print("  Battery "); debugPtr->print(batteryVolts); debugPtr->println("V");
  }
  haveNewPowerSwitchStatus = true;
  return true;
}
