/*
  UART_Client.cpp
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/
#include "BLEDevice.h"
#include <SafeString.h>
#include <millisDelay.h>
#include "UART_Client.h"
#include "BLE_UART_UUIDs.h"


static Print* debugPtr = NULL;
void setUART_ClientDebugOut(Print* outPtr) {
  debugPtr = outPtr;
}

void (*msgParserPtr)(SafeString &sfMsg);

void setMsgParser(void (*_msgParser)(SafeString &sfMsg)) {
  msgParserPtr = _msgParser;
}

static volatile bool connected = false;
//static volatile bool waitingForResponse = false;
static char deviceMsg[MAX_PFOD_MSG_LEN]; // max size of pfodDevice return 1023+ null
static BLEClient*  connectedClientPtr  = BLEDevice::createClient();

static BLERemoteCharacteristic* pRemoteCharacteristic_TX; // read from Server
static BLERemoteCharacteristic* pRemoteCharacteristic_RX; // write to Server

void UART_Client_disconnect() {
  if (connected) {
    connectedClientPtr->disconnect();
    connected = false;
  }
}

bool UART_Client_isConnected() {
  return connected;
}

void sendMsg(const char*msg) {
  if ((connected)) {// && (!isWaitingForResponse())) {
    cSF(sfMsg, 20);
    sfMsg.readFrom(msg); // only read first 20 chars
    if (debugPtr) {
      debugPtr->print("Sending \"");
      debugPtr->print(sfMsg);
      debugPtr->println("\"");
    }
    // Set the characteristic's value to be the array of bytes that is actually a string.
    pRemoteCharacteristic_RX->writeValue(sfMsg.c_str(), sfMsg.length());
    //waitingForResponse = true;
  }
  // else ignore??
}

// NOTE: this is run NOT on the loop() thread so use volatiles to communicate results/state changes
// NOTE: this discards any raw data only { } msg parsed
static void notifyCallback( BLERemoteCharacteristic * pBLERemoteCharacteristic,   uint8_t* pData,  size_t length,  bool isNotify) {
//  if (debugPtr) {
//    debugPtr->print("notifyCallback '");
//    cSF(sfMsg, length);
//    sfMsg.readFrom((char*)pData, length); // read as much as we can, but only upto the specified length
//    debugPtr->print(sfMsg);
//    debugPtr->print("' of data length ");
//    debugPtr->println(length);
//  }
  if (length == 0) {
    return; // nothing to do
  }
  cSFA(sfDeviceMsg, deviceMsg);

  // if sfDeviceMsg isFull have found staring {
  if (sfDeviceMsg.isFull()) { // data over flow just read looking for closing }
    cSF(sfResponse, 25); // >max length 20;
    sfResponse.readFrom((char*)pData, length); // read as much as we can, but only upto the specified length
    int idx = sfDeviceMsg.indexOf('}');
    if (idx >= 0) { // not found keep going
      if (msgParserPtr) {
        (*msgParserPtr)(sfDeviceMsg);
      }
      sfDeviceMsg.clear();
    }
    return;
  }

  sfDeviceMsg.readFrom((char*)pData, length); // read as much as we can, but only upto the specified length
  if (sfDeviceMsg.isEmpty()) { // look for opening {
    int idx = sfDeviceMsg.indexOf('{');
    if (idx < 0) { // not found discard
      sfDeviceMsg.clear();
    } else {
      sfDeviceMsg.removeBefore(idx);
      if (debugPtr) {
        //debugPtr->println("Found starting {");
      }
    }
  }
  int idx = sfDeviceMsg.indexOf('}');
  if (idx >= 0) { // not found keep going
    // got whole msg
    if (debugPtr) {
      //     debugPtr->print("Got response:");     debugPtr->println(sfDeviceMsg);
    }
    if (msgParserPtr) {
      (*msgParserPtr)(sfDeviceMsg);
    }
    sfDeviceMsg.clear();
  } else if (sfDeviceMsg.isFull()) {
    if (debugPtr) {
      debugPtr->print("Got maximum response length:");     debugPtr->println(sfDeviceMsg);
    }
  }
}

// NOTE: this is run NOT on the loop() thread so use volatiles to communicate results/state changes
class MyClientCallback : public BLEClientCallbacks {
    void onConnect(BLEClient* pclient) {
      //sfDeviceMsg.clear();
    }

    void onDisconnect(BLEClient* pclient) {
      connected = false;
      if (debugPtr) {
        debugPtr->println("onDisconnect");
      }
    }
};

bool connectTo(BLEAdvertisedDevice* myDevice) {
  if (!myDevice) {
    if (debugPtr) {
      debugPtr->print("device arg is NULL ");
    }
    return false;
  }
  if (connectedClientPtr->isConnected()) {
    if (debugPtr) {
      debugPtr->print("already connected ");
    }
    return false;
  }

  if (debugPtr) {
    debugPtr->print("Forming a connection to ");
    debugPtr->println(myDevice->getAddress().toString().c_str());
  }

  connectedClientPtr->setClientCallbacks(new MyClientCallback());

  // Connect to the remove BLE Server.
  connectedClientPtr->connect(myDevice);  // if you pass BLEAdvertisedDevice instead of address, it will be recognized type of peer device address (public or private)
  if (debugPtr) {
    debugPtr->println(" - Connected to server");
  }
  // connectedClientPtr->setMTU(517); //set client to request maximum MTU from server 
  // (default is 23 otherwise)

  // Obtain a reference to the service we are after in the remote BLE server.
  BLERemoteService* pRemoteService = connectedClientPtr->getService(serviceUUID);
  if (pRemoteService == nullptr) {
    if (debugPtr) {
      debugPtr->print("Failed to find our service UUID: ");
      debugPtr->println(serviceUUID.toString().c_str());
    }
    connectedClientPtr->disconnect();
    //doScan = true;
    return false;
  }
  if (debugPtr) {
    debugPtr->println(" - Found our service");
  }


  // Obtain a reference to the characteristic in the service of the remote BLE server.
  pRemoteCharacteristic_TX = pRemoteService->getCharacteristic(charUUID_TX);
  if (pRemoteCharacteristic_TX == nullptr) {
    if (debugPtr) {
      debugPtr->print("Failed to find our characteristic UUID: ");
      debugPtr->println(charUUID_TX.toString().c_str());
    }
    connectedClientPtr->disconnect();
    return false;
  }
  if (debugPtr) {
    debugPtr->println(" - Found our TX characteristic");
  }

  // Read the value of the characteristic.
  if (pRemoteCharacteristic_TX->canRead()) {
    std::string value = pRemoteCharacteristic_TX->readValue();
    if (debugPtr) {
      debugPtr->print("The characteristic value was: ");
      debugPtr->println(value.c_str());
    }
  }

  if (pRemoteCharacteristic_TX->canNotify()) {
    pRemoteCharacteristic_TX->registerForNotify(notifyCallback);
  }

  // Obtain a reference to the characteristic in the service of the remote BLE server.
  pRemoteCharacteristic_RX = pRemoteService->getCharacteristic(charUUID_RX);
  if (pRemoteCharacteristic_RX == nullptr) {
    if (debugPtr) {
      debugPtr->print("Failed to find our characteristic UUID: ");
      debugPtr->println(charUUID_RX.toString().c_str());
    }
    connectedClientPtr->disconnect();
    return false;
  }
  if (debugPtr) {
    debugPtr->println(" - Found our RX characteristic");
  }

  // Read the value of the characteristic.
  if (pRemoteCharacteristic_RX->canRead()) {
    std::string value = pRemoteCharacteristic_RX->readValue();
    if (debugPtr) {
      debugPtr->print("The characteristic value was: ");
      debugPtr->println(value.c_str());
    }
  }

  if (pRemoteCharacteristic_RX->canNotify()) {
    pRemoteCharacteristic_RX->registerForNotify(notifyCallback);
  }

  connected = true;
  cSFA(sfDeviceMsg, deviceMsg);
  sfDeviceMsg.clear(); // clear it

  return true;
}
