/*
  WiFiLuxPowerSwitch.ino
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#define ESP32_C3_0_42LCD
// comment to the above define to remove the LCD code used by the 01Space ESP32-C3-0.42LCD https://github.com/01Space/ESP32-C3-0.42LCD
// needs the U8g2 by Oliver V2.33.15 library installed from the Arduino library manager
// The code can be left in for all boards that are not using I2C/Wire for something else.

// NOTE the 01Space ESP32-C3-0.42LCD must be compiled with the board setting Adafruit ESP32-C3 QT PY in order to get Serial out
// for Adafruit ESP32-C3 QT PY use the Adafruit ESP32-C3 QT PY board setting
// Use ESP32C3 Dev Module for the Sparkfun ESP32-C3 (https://www.sparkfun.com/products/18036) did not have any problems

// Using ESP32 based board programmed via Arduino IDE
// follow the steps given on http://www.forward.com.au/pfod/ESP32/index.html to install ESP32 support for Arduino IDE
// The code compiled with ESP32 V2.0.6 board support.  Other version will their own bugs

// You need to modify the WLAN_SSID, WLAN_PASS settings below
// to match your network settings

#include <pfodEEPROM.h>
#include <WiFi.h>
#include "Scanner.h"
#include <SafeString.h>
#include <BufferedOutput.h>
#include <millisDelay.h>
#include "BLE_UART_UUIDs.h"
#include "UART_Client.h"
#include "PowerSwitchMsgParser.h"
#include "ParseSensor.h"

createBufferedOutput(outBuffer, 1024, DROP_IF_FULL);

//#define DEBUG
static Stream* debugPtr = NULL;

#ifdef ESP32_C3_0_42LCD
#include <U8g2lib.h>
U8G2_SSD1306_72X40_ER_F_HW_I2C u8g2(U8G2_R0, /* reset=*/ U8X8_PIN_NONE);   // EastRising 0.42" OLED
#include <Wire.h>
#define SDA_PIN 5
#define SCL_PIN 6
#endif


// install pfodParser library
// download the zip file from http://www.forward.com.au/pfod/pfodParserLibraries/index.html
#include <pfodESPBufferedClient.h>
#include <pfodSecurity.h>
int swap01(int); // method prototype for slider end swaps

pfodSecurity parser("V1"); // create a parser with menu version string to handle the pfod messages
pfodESPBufferedClient bufferedClient;

static const char WL_1[] = "WL_1"; // the lux sensor to drive this switch WL_ devices have temp(degC),RH(%),AirPressure(hPa),Lux

#define WLAN_SSID       "ssid"        // cannot be longer than 32 characters!
#define WLAN_PASS       "password"

const int portNo = 4989; // What TCP port to listen on for connections.
const char staticIP[] = "10.1.1.76";  // set this the static IP you want, e.g. "10.1.1.200" or leave it as "" for DHCP. DHCP is not recommended.

// add your pfod Password here for 128bit security
// eg "b0Ux9akSiwKkwCtcnjTnpWp" but generate your own key, "" means no pfod password
#define pfodSecurityCode ""
// see http://www.forward.com.au/pfod/ArduinoWiFi_simple_pfodDevice/index.html for more information and an example
// and QR image key generator.

WiFiServer server(portNo);
WiFiClient client;
WiFiClient* clientPtr; // non-null if have connected client

static bool PowerOn = false;
static int minVoltsCount = 1023 * 24 / 30; // 2.4V // 2.4V is the minimum operating voltage for the PowerSwitch to turn ON (either relay or low power fet versions)
static int cmd_A_var = 0; // name the variable for 'Switch Power On'  0= 1=
static int cmd_B_var = 0; // name the variable for 'Switch Power Off'  0= 1=
static int adcReading = 0; // for PowerSwitch battery level

static millisDelay connectionTimeout;
static const unsigned long CONNECTION_TIMEOUT_MS = 10000; // 10sec

millisDelay refreshTimer;
static const unsigned long REFRESH_TIME_MS = 5000; // check setting every 5sec

static millisDelay powerSwitchStatusTimer;
static const unsigned long POWER_SWITCH_STATUS_TIMER_MS = REFRESH_TIME_MS + CONNECTION_TIMEOUT_MS; // should have new status by this time
static millisDelay luxReadingTimer;
static const unsigned long LUX_READING_TIMER_MS = 120000; // 120sec, get new broadcast every 100sec

static const char noLuxData[] = "---- lx";
static const char noPowerSwitchData[] = "---- V";
cSF(sfPowerSwitchStatus, 10); // actually 8 char is about as much as can be seen
cSF(sfLuxReading, 10); // actually 8 char is about as much as can be seen
cSF(sfBatteryVolts, 10); // actually 4 chars
static bool haveSwitchStatus = false;
static bool haveLuxSensorReading = false;

static volatile bool wasConnected = false;
static volatile bool waitingForResponse = false;

static cSF(sfNextCmd, 20);

static const float SUNSET_SUNRISE_LX = 10.0; // divider between sunrise/sunset

static millisDelay sunset_sunrise_delay;  // once the lux measurement triggers a change ignore other measurement for the next 10 mins
static unsigned long SUNSET_SUNRISE_DELAY_MS = 10UL * 60 * 1000; // 10min
static float luxReading  = -100.0; // not set yet
static bool initializeLuxPowerOn = true; // first time only
static bool luxPowerOn = true; // this keeps track of the power setting triggered by the lux reading, not actually used here

static volatile bool haveNewLuxReading = false;

static char deviceAddress[] = "F5:FA:14:1C:85:B7";  // either upper or lower case OK here
static volatile BLEAdvertisedDevice* myDevice = NULL;


// wdt to reboot esp32 if BLE locks up
static const unsigned long wdtTimeout = 20000;  //20sec time in ms to trigger the watchdog
static hw_timer_t *timer = NULL;

void ARDUINO_ISR_ATTR resetModule() {
  if (debugPtr) {
    debugPtr->println("Reboot");
    debugPtr->flush();
  }
  esp_restart();
}


// call this to send next cmd
// only called from loop() never from callbacks
void setNextCmd(const char* nextCmd) {
  sfNextCmd.clear();
  sfNextCmd.readFrom(nextCmd);
}

// prints to debug but otherwise does nothing if not ESP32_C3_0_42LCD
void updateDisplay() {
  if (debugPtr) {
    debugPtr->println(sfLuxReading);
    debugPtr->println(sfPowerSwitchStatus);
  }
#ifdef ESP32_C3_0_42LCD
  u8g2.clearBuffer();
  u8g2.setFlipMode(1);
  u8g2.setFontMode(1); // << transparent ??
  u8g2.setFont(u8g2_font_crox4tb_tf);
  u8g2.setCursor(0, 15);
  u8g2.print(sfLuxReading.c_str()); // e.g. >999 lx
  u8g2.setCursor(0, 37);
  u8g2.print(sfPowerSwitchStatus.c_str()); // e.g. Off 3.0V;
  u8g2.sendBuffer();
#endif
}


/**
   Scan for BLE servers and find the first one that advertises the service we are looking for.
*/
class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
      //      if (debugPtr) {
      //        debugPtr->print("BLE Advertised : ");
      //        debugPtr->println(advertisedDevice.toString().c_str());
      //      }
      BLEAddress address =  advertisedDevice.getAddress();
      std::string addressStr = address.toString();  // << returns address in lower case
      cSFA(sfDeviceAddress, deviceAddress);
      if (sfDeviceAddress.equalsIgnoreCase(addressStr.c_str())) { // handles upper and lower hex
        if (debugPtr) {
          debugPtr->print("Found Power Switch "); debugPtr->print(deviceAddress);
          if (advertisedDevice.haveRSSI()) {
            debugPtr->print(" rssi:"); debugPtr->print(advertisedDevice.getRSSI());
          }
          debugPtr->println();
        }
        // We have found a device, let us now see if it contains the service we are looking for.
        if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(serviceUUID)) {
          myDevice = new BLEAdvertisedDevice(advertisedDevice); //main loop will connect
          // Found our server
        } else {
          if (debugPtr) {
            debugPtr->print("BLE device does not have UART service");   debugPtr->println();
          }
          myDevice = NULL;
        }
        return;
      }
      // else
      if (parseDeviceName(advertisedDevice.getName().c_str())) {
        if (debugPtr) {
          debugPtr->print("Found Sensor: ");
          debugPtr->print(advertisedDevice.getName().c_str());
          if (advertisedDevice.haveRSSI()) {
            debugPtr->print(" rssi:"); debugPtr->print(advertisedDevice.getRSSI());
          }
          debugPtr->println();
        }
        haveNewLuxReading = true; // loop() can now pickup new reading
        return;
      }
    } // onResult
}; // MyAdvertisedDeviceCallbacks


// on call back thread
void copyMsg(SafeString &sfDeviceMsg) {
  const char *msg = sfDeviceMsg.c_str();
  volatile char *psMsg = powerSwitchMsg;
  while (*msg) {
    *psMsg = *msg;
    psMsg++;
    msg++;
  }
  *psMsg = '\0';
  //  if (debugPtr) {
  //    debugPtr->println((const char*)powerSwitchMsg);
  //  }
}

void processLuxReading() {
  if (sunset_sunrise_delay.justFinished()) {
    // do nothing here
  }
  if (!UART_Client_isConnected()) {
    return; // can not sent cmd usually only get this on startup if scan lux reading before scanning power switch and connecting
  }
  if ( luxReading < 0) {
    return; // not set yet
  }
  if (( luxReading > SUNSET_SUNRISE_LX) && (!sunset_sunrise_delay.isRunning()) && (luxPowerOn || initializeLuxPowerOn)) {
    initializeLuxPowerOn = false;
    luxPowerOn = false;
    sunset_sunrise_delay.start(SUNSET_SUNRISE_DELAY_MS);
    setNextCmd("{B`1}");
  }
  if (( luxReading < SUNSET_SUNRISE_LX) && (!sunset_sunrise_delay.isRunning()) && ((!luxPowerOn) || initializeLuxPowerOn)) {
    initializeLuxPowerOn = false;
    luxPowerOn = true;
    sunset_sunrise_delay.start(SUNSET_SUNRISE_DELAY_MS);
    setNextCmd("{A`1}");
  }
}


unsigned long plot_msOffset = 0; // set by {@} response
bool clearPlot = false; // set by the {@} response code

// the setup routine runs once on reset:
void setup() {
  // Explicitly set the ESP32 to be a WiFi-client, otherwise, it by default,
  // would try to act as both a client and an access-point and could cause
  // network-issues with your other WiFi-devices on your WiFi-network.
  WiFi.mode(WIFI_STA);
  EEPROM.begin(512);  // only use 20bytes for pfodSecurity but reserve 512 (pfodWifiConfig uses more)
#ifdef DEBUG
  Serial.begin(115200);
  Serial.println();
  outBuffer.connect(Serial); // always even if not debugging
  debugPtr = &outBuffer;
  // these statements enable debug output in the respective modules
  //setPowerSwitchMsgParserDebugOut(debugPtr);
  //setParseSensorDebugOut(debugPtr);
  //setPowerSwitchMsgParserDebugOut(debugPtr);
  //setScannerDebugOut(debugPtr);
  //setUART_ClientDebugOut(debugPtr);
#endif

#ifdef ESP32_C3_0_42LCD
  Wire.begin(SDA_PIN, SCL_PIN);
  u8g2.begin();
#endif

  /* Initialise wifi module */
  IPAddress ip;
  if ((*staticIP != '\0') && (ip.fromString(staticIP))) {
    IPAddress gateway(ip[0], ip[1], ip[2], 1); // set gatway to ... 1
#ifdef DEBUG
    Serial.print(F("Setting gateway to: "));
    Serial.println(gateway);
#endif
    IPAddress subnet(255, 255, 255, 0);
    WiFi.config(ip, gateway, subnet);
  }
  WiFi.begin(WLAN_SSID, WLAN_PASS);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
#ifdef DEBUG
    Serial.print(".");
#endif
  }
#ifdef DEBUG
  Serial.println();
  Serial.println("WiFi connected");
#endif

  // Start the server
  server.begin();
#ifdef DEBUG
  Serial.println("Server started");
#endif

  // Print the IP address
#ifdef DEBUG
  Serial.println(WiFi.localIP());
#endif

#ifdef ESP32_C3_0_42LCD
  Wire.begin(SDA_PIN, SCL_PIN);
  u8g2.begin();
#endif

  // start missing data timers
  powerSwitchStatusTimer.start(POWER_SWITCH_STATUS_TIMER_MS);
  luxReadingTimer.start(LUX_READING_TIMER_MS);
  //set initial screen display
  sfPowerSwitchStatus =  noPowerSwitchData;
  sfLuxReading =  noLuxData;
  sfBatteryVolts = noPowerSwitchData;
  updateDisplay(); // initial display

  setLuxSensorName(WL_1);
  setMsgParser(copyMsg); // handle BLE device responses copy to volatile char[] for later pasrsing on loop()
  startScanner(new MyAdvertisedDeviceCallbacks() );



  // setup and start wdt
  timer = timerBegin(0, 80, true);                  //timer 0, div 80
  timerAttachInterrupt(timer, &resetModule, true);  //attach callback
  timerAlarmWrite(timer, wdtTimeout * 1000, false); //set time in us
  timerAlarmEnable(timer);                          //enable interrupt

  // <<<<<<<<< Your extra setup code goes here
}

void togglePowerSwitch() {
  if (PowerOn) { // turn off
    setNextCmd("{B`1}");
    PowerOn = false;
  } else { // power off turn on
    if (adcReading >= minVoltsCount) { // <<<<<<<<< check supply volts
      setNextCmd("{A`1}");
      PowerOn = true;
    } else {
      // ignore this ON cmd if low volts
    }
  }
}

// the loop routine runs over and over again forever:
void loop() {
  timerWrite(timer, 0); //reset timer (feed watchdog)
  if (debugPtr) {
    outBuffer.nextByteOut(); // always push out buffered debug
  }

  restartScanner(); // if necessary

  if (parsePowerSwitchMsg()) { // got a msg back from the Power Switch
    waitingForResponse = false;
    haveSwitchStatus = true;
    powerSwitchStatusTimer.restart();
    adcReading = getADCreading();
    sfBatteryVolts.clear();
    sfBatteryVolts.print(((float)adcReading) / 1023.0 * 3.0, 2); // one decimal place
    sfBatteryVolts += "V";

    sfPowerSwitchStatus = isPowerSwitchOn() ? "On  " : "0ff ";
    sfPowerSwitchStatus += sfBatteryVolts;
    updateDisplay();
  }

  if (haveNewLuxReading) {
    haveNewLuxReading = false;
    haveLuxSensorReading = true;
    luxReading = getLuxReading();
    // do other stuff here with this reading
    luxReadingTimer.restart(); // check for sensor out of range / off
    if (luxReading < 0) {
      sfLuxReading = noLuxData;
    } else if (luxReading > 999) {
      sfLuxReading = ">999 lx";
    } else {
      // limit luxReading to 3digits + .
      sfLuxReading.clear();
      sfLuxReading.print(luxReading, 2, 4);
      sfLuxReading += " lx";
    }
    updateDisplay();
  }


  bool connected = UART_Client_isConnected();

  if ((!connected) && myDevice) {
    sfPowerSwitchStatus = noPowerSwitchData; // no status
    updateDisplay();
    if (debugPtr) {
      debugPtr->println("Connecting ... ");
    }
    if (connectTo((BLEAdvertisedDevice*)myDevice)) { // does nothing if myDevice is NULL OR if already connected
      // this call is blocking and sometimes locks up in which case the ESP32 will reboot
      if (debugPtr) {
        debugPtr->println("*** Connected to the BLE Server.");
      }
      wasConnected = true;
      waitingForResponse = false;
      setNextCmd("{.}"); // get current state
    } else {
      if (debugPtr) {
        debugPtr->println("!!! Failed to connect to the BLE server.");
      }
    }
  }

  if ((!connectionTimeout.isRunning()) && (!refreshTimer.isRunning())) {
    refreshTimer.start(REFRESH_TIME_MS);
  }

  // If we are connected to a peer BLE Server, update the characteristic each time we are reached
  // with the current time since boot.
  if (refreshTimer.justFinished()) { // always do this test at the top level in the loop
    if ((connected) && sfNextCmd.isEmpty()) { // do not override any cmd from WiFi input
      setNextCmd("{.}");
    }
  }

  processLuxReading(); // do this AFTER refresh setNextCmd so the cmd can be overridded by the lux reading control

  // if have a cmd to send and not waiting for response send it
  if ((!sfNextCmd.isEmpty()) && (!waitingForResponse)  && connected) {
    waitingForResponse = true;  // set this here because short responses come back very fast and interleave between statements below
    if (debugPtr) {
      debugPtr->print("Sending '"); debugPtr->print(sfNextCmd);  debugPtr->println("'");
    }
    sendMsg(sfNextCmd.c_str());
    sfNextCmd.clear();
    refreshTimer.start(REFRESH_TIME_MS);
    //    connectionTimeout.start(CONNECTION_TIMEOUT_MS);
  }

  if (clientPtr && (!clientPtr->connected())) {
    closeConnection(parser.getPfodAppStream());
  }
  if (server.hasClient()) { // new connection
    if (!clientPtr) { // no existing connection
      client = server.available(); // get any new client
      clientPtr = &client;
      // have connected client
      parser.connect(bufferedClient.connect(clientPtr), F(pfodSecurityCode)); // sets new io stream to read from and write to
      EEPROM.commit(); // does nothing if nothing to do
    } else { // already have client so just stop the new one
      WiFiClient newClient = server.available(); // get any new client
      newClient.stop();
    }
  }

  if (clientPtr) {
    uint8_t cmd = parser.parse(); // parse incoming data from connection
    // parser returns non-zero when a pfod command is fully parsed
    if (cmd != 0) { // have parsed a complete msg { to }
      uint8_t* pfodFirstArg = parser.getFirstArg(); // may point to \0 if no arguments in this msg.
      pfod_MAYBE_UNUSED(pfodFirstArg); // may not be used, just suppress warning
      long pfodLongRtn; // used for parsing long return arguments, if any
      pfod_MAYBE_UNUSED(pfodLongRtn); // may not be used, just suppress warning
      if ('.' == cmd) {
        // pfodApp has connected and sent {.} , it is asking for the main menu
        if (!parser.isRefresh()) {
          sendMainMenu(); // send back the menu designed
        } else {
          sendMainMenuUpdate(); // menu is cached just send update
        }

        // handle {@} request
      } else if ('@' == cmd) { // pfodApp requested 'current' time
        plot_msOffset = millis(); // capture current millis as offset rawdata timestamps
        clearPlot = true; // clear plot on reconnect as have new plot_msOffset
        parser.print(F("{@`0}")); // return `0 as 'current' raw data milliseconds

        // now handle commands returned from button/sliders
      } else if ('A' == cmd) { // user moved slider -- 'Switch Power On'
        // in the main Menu of Power Switch
        // set output based on slider 0= 1=
        if (adcReading >= minVoltsCount) { // <<<<<<<<< check supply volts from last reading
          parser.parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
          cmd_A_var = (int)pfodLongRtn; // set variable
          if (cmd_A_var == 1) { // send {A`1}
            setNextCmd("{A`1}");
            PowerOn = true;
            setPowerIsOn(PowerOn);
            cmd_A_var = 0;
          }
        } else {
          // ignore this ON cmd if low volts
        }
        sendMainMenuUpdate(); // always send back a pfod msg otherwise pfodApp will disconnect.

      } else if ('B' == cmd) { // user moved slider -- 'Switch Power Off'
        // in the main Menu of Power Switch
        // set output based on slider 0= 1=
        parser.parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
        cmd_B_var = (int)pfodLongRtn; // set variable
        if (cmd_B_var == 1) {
          setNextCmd("{B`1}");
          PowerOn = false;
          setPowerIsOn(PowerOn);
          cmd_B_var = 0;
        }
        sendMainMenuUpdate(); // always send back a pfod msg otherwise pfodApp will disconnect.

        //    } else if('C'==cmd) { // this is a label. pfodApp NEVER sends this cmd -- 'Battery'
        //      // in the main Menu of Power Switch

        // add toggle cmd for single button remote control
      } else if ('T' == cmd) { // toggle cmd this is NOT displayed by pfodApp
        togglePowerSwitch();
        PowerOn = !PowerOn;
        sendMainMenuUpdate(); // always send back a pfod msg otherwise pfodApp will disconnect.

        // now handle commands returned from button/sliders
      } else if ('!' == cmd) {
        // CloseConnection command
        closeConnection(parser.getPfodAppStream());
      } else {
        // unknown command
        parser.print(F("{}")); // always send back a pfod msg otherwise pfodApp will disconnect.
      }
    }
  }

  //  <<<<<<<<<<<  Your other loop() code goes here
  if (powerSwitchStatusTimer.justFinished()) {
    powerSwitchStatusTimer.restart();
    sfPowerSwitchStatus = noPowerSwitchData; // no status
    haveSwitchStatus = false;
    updateDisplay();
  }
  if (luxReadingTimer.justFinished()) {
    luxReadingTimer.restart();
    sfLuxReading = noLuxData; // no status
    haveLuxSensorReading = false;
    updateDisplay();
  }

}

void closeConnection(Stream *io) {
  (void)(io); // unused
  // add any special code here to force connection to be dropped
  parser.closeConnection(); // nulls io stream
  bufferedClient.stop(); // clears client reference
  clientPtr->stop();
  clientPtr = NULL;
}

void sendMainMenu() {
  // !! Remember to change the parser version string
  //    every time you edit this method
  parser.print(F("{,"));  // start a Menu screen pfod message
  // send menu background, format, prompt, refresh and version
  parser.print(F("<bg n><b><+8>~Power is\n"));
  if (haveSwitchStatus) {
    parser.print(isPowerSwitchOn() ? "ON" : "OFF");
  } else {
    parser.print("?");
  }
  parser.print("`5000"); // request refresh menu every 10sec
  parser.sendVersion(); // send the menu version
  // send menu items
  parser.print(F("|A<bg s><b><+6><r>"));
  parser.print('`');
  parser.print(cmd_A_var); // output the current state 0 Low or 1 High
  parser.print(F("~Switch Power\nOn~~\\~t"));
  // Note the \\ inside the "'s to send \ ...
  parser.print(F("|B<bg s><b><+6>"));
  parser.print('`');
  parser.print(cmd_B_var); // output the current state 0 Low or 1 High
  parser.print(F("~Switch Power\nOff~~\\~t"));
  // Note the \\ inside the "'s to send \ ...
  parser.print(F("|!C~Battery "));
  parser.print(sfBatteryVolts);
  parser.print("\nLux ");
  if (haveLuxSensorReading) {
    if (luxReading < 99) {
      parser.print(luxReading, 1);
    } else {
      parser.print(luxReading, 0);
    }
  } else {
    parser.print("----");
  }
  parser.print(" lx");
  parser.print(F("}"));  // close pfod message
}

void sendMainMenuUpdate() {
  parser.print(F("{;"));  // start an Update Menu pfod message
  parser.print(F("<bg n><b><+8>~Power is\n"));
  if (haveSwitchStatus) {
    parser.print(isPowerSwitchOn() ? "ON" : "OFF");
  } else {
    parser.print("?");
  }
  parser.print("`5000"); // request refresh menu every 10 sec
  // send menu items
  parser.print(F("|A"));
  parser.print('`');
  parser.print(cmd_A_var); // output the current state 0 Low or 1 High
  parser.print(F("|B"));
  parser.print('`');
  parser.print(cmd_B_var); // output the current state 0 Low or 1 High
  parser.print(F("|!C~Battery "));
  parser.print(sfBatteryVolts);
  parser.print("\nLux ");
  if (haveLuxSensorReading) {
    if (luxReading < 99) {
      parser.print(luxReading, 1);
    } else {
      parser.print(luxReading, 0);
    }
  } else {
    parser.print("----");
  }
  parser.print(" lx");
  parser.print(F("}"));  // close pfod message
  // ============ end of menu ===========
}

int swap01(int in) {
  return (in == 0) ? 1 : 0;
}
// ============= end generated code =========
